@@ -815,7 +815,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
         case 'G': {
             /* Cursor horizontal absolute */
             int col = min(vt_param_get(term, 0, 1), term->cols) - 1;
-            term_cursor_to(term, term->grid->cursor.point.row, col);
+            term_cursor_col(term, col);
             break;
         }
 
@@ -1206,8 +1206,10 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
                 if (width >= 0 && height >= 0) {
                     char reply[64];
-                    size_t n = xsnprintf(reply, sizeof(reply), "\033[4;%d;%dt",
-                             height / term->scale, width / term->scale);
+                    size_t n = xsnprintf(
+                        reply, sizeof(reply), "\033[4;%d;%dt",
+                        (int)round(height / term->scale),
+                        (int)(width / term->scale));
                     term_to_slave(term, reply, n);
                 }
                 break;
@@ -1229,9 +1231,10 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
             case 16: { /* report cell size in pixels */
                 char reply[64];
-                size_t n = xsnprintf(reply, sizeof(reply), "\033[6;%d;%dt",
-                         term->cell_height / term->scale,
-                         term->cell_width / term->scale);
+                size_t n = xsnprintf(
+                    reply, sizeof(reply), "\033[6;%d;%dt",
+                    (int)round(term->cell_height / term->scale),
+                    (int)round(term->cell_width / term->scale));
                 term_to_slave(term, reply, n);
                 break;
             }
@@ -1247,9 +1250,10 @@ csi_dispatch(struct terminal *term, uint8_t final)
             case 19: { /* report screen size in chars */
                 tll_foreach(term->window->on_outputs, it) {
                     char reply[64];
-                    size_t n = xsnprintf(reply, sizeof(reply), "\033[9;%d;%dt",
-                             it->item->dim.px_real.height / term->cell_height / term->scale,
-                             it->item->dim.px_real.width / term->cell_width / term->scale);
+                    size_t n = xsnprintf(
+                        reply, sizeof(reply), "\033[9;%d;%dt",
+                        (int)round(it->item->dim.px_real.height / term->cell_height / term->scale),
+                        (int)round(it->item->dim.px_real.width / term->cell_width / term->scale));
                     term_to_slave(term, reply, n);
                     break;
                 }
@@ -1398,6 +1402,39 @@ csi_dispatch(struct terminal *term, uint8_t final)
             break;
         }
 
+        case 'm': {
+            int resource = vt_param_get(term, 0, 0);
+            int value = -1;
+
+            switch (resource) {
+            case 0:  /* modifyKeyboard */
+                value = 0;
+                break;
+
+            case 1:  /* modifyCursorKeys */
+            case 2:  /* modifyFunctionKeys */
+                value = 1;
+                break;
+
+            case 4:  /* modifyOtherKeys */
+                value = term->modify_other_keys_2 ? 2 : 1;
+                break;
+
+            default:
+                LOG_WARN("XTQMODKEYS: invalid resource '%d' in '%s'",
+                         resource, csi_as_string(term, final, -1));
+                break;
+            }
+
+            if (value >= 0) {
+                char reply[16] = {0};
+                int chars = snprintf(reply, sizeof(reply),
+                                     "\033[>%d;%dm", resource, value);
+                term_to_slave(term, reply, chars);
+            }
+            break;
+        }
+
         case 'u': {
             enum kitty_kbd_flags flags =
                 term->grid->kitty_kbd.flags[term->grid->kitty_kbd.idx];
@@ -1489,7 +1526,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
                     break;
 
                 default:
-                    LOG_WARN("invalid resource %d in %s",
+                    LOG_WARN("XTMODKEYS: invalid resource '%d' in '%s'",
                              resource, csi_as_string(term, final, -1));
                     break;
                 }
