@@ -14,6 +14,10 @@
 #include <wayland-cursor.h>
 #include <xkbcommon/xkbcommon-compose.h>
 
+#if defined(HAVE_CURSOR_SHAPE)
+#include <cursor-shape-v1.h>
+#endif
+
 #include <tllist.h>
 
 #define LOG_MODULE "wayland"
@@ -32,12 +36,12 @@
 #include "xmalloc.h"
 
 static void
-csd_reload_font(struct wl_window *win, int old_scale)
+csd_reload_font(struct wl_window *win, float old_scale)
 {
     struct terminal *term = win->term;
     const struct config *conf = term->conf;
 
-    const int scale = term->scale;
+    const float scale = term->scale;
 
     bool enable_csd = win->csd_mode == CSD_YES && !win->is_fullscreen;
     if (!enable_csd)
@@ -52,10 +56,10 @@ csd_reload_font(struct wl_window *win, int old_scale)
         patterns[i] = conf->csd.font.arr[i].pattern;
 
     char pixelsize[32];
-    snprintf(pixelsize, sizeof(pixelsize),
-             "pixelsize=%u", conf->csd.title_height * scale * 1 / 2);
+    snprintf(pixelsize, sizeof(pixelsize), "pixelsize=%u",
+             (int)round(conf->csd.title_height * scale * 1 / 2));
 
-    LOG_DBG("loading CSD font \"%s:%s\" (old-scale=%d, scale=%d)",
+    LOG_DBG("loading CSD font \"%s:%s\" (old-scale=%.2f, scale=%.2f)",
             patterns[0], pixelsize, old_scale, scale);
 
     win->csd.font = fcft_from_name(conf->csd.font.count, patterns, pixelsize);
@@ -74,12 +78,12 @@ csd_instantiate(struct wl_window *win)
 
     for (size_t i = CSD_SURF_MINIMIZE; i < CSD_SURF_COUNT; i++) {
         bool ret = wayl_win_subsurface_new_with_custom_parent(
-            win, win->csd.surface[CSD_SURF_TITLE].surf, &win->csd.surface[i],
+            win, win->csd.surface[CSD_SURF_TITLE].surface.surf, &win->csd.surface[i],
             true);
         xassert(ret);
     }
 
-    csd_reload_font(win, -1);
+    csd_reload_font(win, -1.);
 }
 
 static void
@@ -187,8 +191,12 @@ seat_destroy(struct seat *seat)
 
     if (seat->pointer.theme != NULL)
         wl_cursor_theme_destroy(seat->pointer.theme);
-    if (seat->pointer.surface != NULL)
-        wl_surface_destroy(seat->pointer.surface);
+    if (seat->pointer.surface.surf != NULL)
+        wl_surface_destroy(seat->pointer.surface.surf);
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (seat->pointer.surface.viewport != NULL)
+        wp_viewport_destroy(seat->pointer.surface.viewport);
+#endif
     if (seat->pointer.xcursor_callback != NULL)
         wl_callback_destroy(seat->pointer.xcursor_callback);
 
@@ -205,10 +213,17 @@ seat_destroy(struct seat *seat)
     if (seat->data_device != NULL)
         wl_data_device_release(seat->data_device);
 
+#if defined(HAVE_CURSOR_SHAPE)
+    if (seat->pointer.shape_device != NULL)
+        wp_cursor_shape_device_v1_destroy(seat->pointer.shape_device);
+#endif
+
     if (seat->wl_keyboard != NULL)
         wl_keyboard_release(seat->wl_keyboard);
     if (seat->wl_pointer != NULL)
         wl_pointer_release(seat->wl_pointer);
+    if (seat->wl_touch != NULL)
+        wl_touch_release(seat->wl_touch);
 
 #if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
     if (seat->wl_text_input != NULL)
@@ -221,6 +236,7 @@ seat_destroy(struct seat *seat)
     ime_reset_pending(seat);
     free(seat->clipboard.text);
     free(seat->primary.text);
+    free(seat->pointer.last_custom_xcursor);
     free(seat->name);
 }
 
@@ -270,9 +286,10 @@ seat_handle_capabilities(void *data, struct wl_seat *wl_seat,
     struct seat *seat = data;
     xassert(seat->wl_seat == wl_seat);
 
-    LOG_DBG("%s: keyboard=%s, pointer=%s", seat->name,
+    LOG_DBG("%s: keyboard=%s, pointer=%s, touch=%s", seat->name,
             (caps & WL_SEAT_CAPABILITY_KEYBOARD) ? "yes" : "no",
-            (caps & WL_SEAT_CAPABILITY_POINTER) ? "yes" : "no");
+            (caps & WL_SEAT_CAPABILITY_POINTER) ? "yes" : "no",
+            (caps & WL_SEAT_CAPABILITY_TOUCH) ? "yes" : "no");
 
     if (caps & WL_SEAT_CAPABILITY_KEYBOARD) {
         if (seat->wl_keyboard == NULL) {
@@ -288,31 +305,81 @@ seat_handle_capabilities(void *data, struct wl_seat *wl_seat,
 
     if (caps & WL_SEAT_CAPABILITY_POINTER) {
         if (seat->wl_pointer == NULL) {
-            xassert(seat->pointer.surface == NULL);
-            seat->pointer.surface = wl_compositor_create_surface(seat->wayl->compositor);
+            xassert(seat->pointer.surface.surf == NULL);
+            seat->pointer.surface.surf =
+                wl_compositor_create_surface(seat->wayl->compositor);
 
-            if (seat->pointer.surface == NULL) {
+            if (seat->pointer.surface.surf == NULL) {
                 LOG_ERR("%s: failed to create pointer surface", seat->name);
                 return;
             }
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+            xassert(seat->pointer.surface.viewport == NULL);
+            seat->pointer.surface.viewport = wp_viewporter_get_viewport(
+                seat->wayl->viewporter, seat->pointer.surface.surf);
+
+            if (seat->pointer.surface.viewport == NULL) {
+                LOG_ERR("%s: failed to create pointer viewport", seat->name);
+                wl_surface_destroy(seat->pointer.surface.surf);
+                seat->pointer.surface.surf = NULL;
+                return;
+            }
+#endif
+
             seat->wl_pointer = wl_seat_get_pointer(wl_seat);
             wl_pointer_add_listener(seat->wl_pointer, &pointer_listener, seat);
+
+#if defined(HAVE_CURSOR_SHAPE)
+            if (seat->wayl->cursor_shape_manager != NULL) {
+                xassert(seat->pointer.shape_device == NULL);
+                seat->pointer.shape_device = wp_cursor_shape_manager_v1_get_pointer(
+                    seat->wayl->cursor_shape_manager, seat->wl_pointer);
+            }
+#endif
         }
     } else {
         if (seat->wl_pointer != NULL) {
+#if defined(HAVE_CURSOR_SHAPE)
+            if (seat->pointer.shape_device != NULL) {
+                wp_cursor_shape_device_v1_destroy(seat->pointer.shape_device);
+                seat->pointer.shape_device = NULL;
+            }
+#endif
+
             wl_pointer_release(seat->wl_pointer);
-            wl_surface_destroy(seat->pointer.surface);
+            wl_surface_destroy(seat->pointer.surface.surf);
+
+#if defined(HAVE_FRACTIONAL_SCALE)
+            wp_viewport_destroy(seat->pointer.surface.viewport);
+            seat->pointer.surface.viewport = NULL;
+#endif
 
             if (seat->pointer.theme != NULL)
                 wl_cursor_theme_destroy(seat->pointer.theme);
 
             seat->wl_pointer = NULL;
-            seat->pointer.surface = NULL;
+            seat->pointer.surface.surf = NULL;
             seat->pointer.theme = NULL;
             seat->pointer.cursor = NULL;
         }
     }
+
+    if (caps & WL_SEAT_CAPABILITY_TOUCH) {
+        if (seat->wl_touch == NULL) {
+            seat->wl_touch = wl_seat_get_touch(wl_seat);
+            wl_touch_add_listener(seat->wl_touch, &touch_listener, seat);
+
+            seat->touch.state = TOUCH_STATE_IDLE;
+        }
+    } else {
+        if (seat->wl_touch != NULL) {
+            wl_touch_release(seat->wl_touch);
+            seat->wl_touch = NULL;
+        }
+
+        seat->touch.state = TOUCH_STATE_INHIBITED;
+    }
 }
 
 static void
@@ -331,15 +398,35 @@ static const struct wl_seat_listener seat_listener = {
 static void
 update_term_for_output_change(struct terminal *term)
 {
-    if (tll_length(term->window->on_outputs) == 0)
-        return;
+    const float old_scale = term->scale;
+    const float logical_width = term->width / term->scale;
+    const float logical_height = term->height / term->scale;
 
-    int old_scale = term->scale;
-
-    render_resize(term, term->width / term->scale, term->height / term->scale);
-    term_font_dpi_changed(term, old_scale);
+    /* Note: order matters! term_update_scale() must come first */
+    bool scale_updated = term_update_scale(term);
+    bool fonts_updated = term_font_dpi_changed(term, old_scale);
     term_font_subpixel_changed(term);
+
     csd_reload_font(term->window, old_scale);
+
+    if (fonts_updated) {
+        /*
+         * If the fonts have been updated, the cell dimensions have
+         * changed. This requires a “forced” resize, since the surface
+         * buffer dimensions may not have been updated (in which case
+         * render_size() normally shortcuts and returns early).
+         */
+        render_resize_force(term, round(logical_width), round(logical_height));
+    }
+
+    else if (scale_updated) {
+        /*
+         * A scale update means the surface buffer dimensions have
+         * been updated, even though the window logical dimensions
+         * haven’t changed.
+         */
+        render_resize(term, round(logical_width), round(logical_height));
+    }
 }
 
 static void
@@ -350,11 +437,6 @@ update_terms_on_monitor(struct monitor *mon)
     tll_foreach(wayl->terms, it) {
         struct terminal *term = it->item;
 
-        if (term->conf->dpi_aware == DPI_AWARE_AUTO) {
-            update_term_for_output_change(term);
-            continue;
-        }
-
         tll_foreach(term->window->on_outputs, it2) {
             if (it2->item == mon) {
                 update_term_for_output_change(term);
@@ -373,6 +455,9 @@ output_update_ppi(struct monitor *mon)
     double x_inches = mon->dim.mm.width * 0.03937008;
     double y_inches = mon->dim.mm.height * 0.03937008;
 
+    const int width = mon->dim.px_real.width;
+    const int height = mon->dim.px_real.height;
+
     mon->ppi.real.x = mon->dim.px_real.width / x_inches;
     mon->ppi.real.y = mon->dim.px_real.height / y_inches;
 
@@ -395,27 +480,36 @@ output_update_ppi(struct monitor *mon)
         break;
     }
 
-    int scaled_width = mon->dim.px_scaled.width;
-    int scaled_height = mon->dim.px_scaled.height;
-
-    if (scaled_width == 0 && scaled_height == 0 && mon->scale > 0) {
-        /* Estimate scaled width/height if none has been provided */
-        scaled_width = mon->dim.px_real.width / mon->scale;
-        scaled_height = mon->dim.px_real.height / mon->scale;
-    }
+    const int scaled_width = mon->dim.px_scaled.width;
+    const int scaled_height = mon->dim.px_scaled.height;
 
     mon->ppi.scaled.x = scaled_width / x_inches;
     mon->ppi.scaled.y = scaled_height / y_inches;
 
-    double px_diag = sqrt(pow(scaled_width, 2) + pow(scaled_height, 2));
-    mon->dpi = px_diag / mon->inch * mon->scale;
+    const double px_diag_physical = sqrt(pow(width, 2) + pow(height, 2));
+    mon->dpi.physical = width == 0 && height == 0
+        ? 96.
+        : px_diag_physical / mon->inch;
 
-    if (mon->dpi > 1000) {
+    const double px_diag_scaled = sqrt(pow(scaled_width, 2) + pow(scaled_height, 2));
+    mon->dpi.scaled = scaled_width == 0 && scaled_height == 0
+        ? 96.
+        : px_diag_scaled / mon->inch * mon->scale;
+
+    if (mon->dpi.physical > 1000) {
         if (mon->name != NULL) {
-            LOG_WARN("%s: DPI=%f is unreasonable, using 96 instead",
-                     mon->name, mon->dpi);
+            LOG_WARN("%s: DPI=%f (physical) is unreasonable, using 96 instead",
+                     mon->name, mon->dpi.physical);
         }
-        mon->dpi = 96;
+        mon->dpi.physical = 96;
+    }
+
+    if (mon->dpi.scaled > 1000) {
+        if (mon->name != NULL) {
+            LOG_WARN("%s: DPI=%f (logical) is unreasonable, using 96 instead",
+                     mon->name, mon->dpi.scaled);
+        }
+        mon->dpi.scaled = 96;
     }
 }
 
@@ -633,6 +727,7 @@ xdg_toplevel_configure(void *data, struct xdg_toplevel *xdg_toplevel,
     bool is_tiled_bottom = false;
     bool is_tiled_left = false;
     bool is_tiled_right = false;
+    bool is_suspended UNUSED = false;
 
 #if defined(LOG_ENABLE_DBG) && LOG_ENABLE_DBG
     char state_str[2048];
@@ -647,29 +742,35 @@ xdg_toplevel_configure(void *data, struct xdg_toplevel *xdg_toplevel,
         [XDG_TOPLEVEL_STATE_TILED_RIGHT] = "tiled:right",
         [XDG_TOPLEVEL_STATE_TILED_TOP] = "tiled:top",
         [XDG_TOPLEVEL_STATE_TILED_BOTTOM] = "tiled:bottom",
+#if defined(XDG_TOPLEVEL_STATE_SUSPENDED_SINCE_VERSION)  /* wayland-protocols >= 1.32 */
+        [XDG_TOPLEVEL_STATE_SUSPENDED] = "suspended",
+#endif
     };
 #endif
 
     enum xdg_toplevel_state *state;
     wl_array_for_each(state, states) {
         switch (*state) {
-        case XDG_TOPLEVEL_STATE_ACTIVATED:    is_activated = true; break;
-        case XDG_TOPLEVEL_STATE_FULLSCREEN:   is_fullscreen = true; break;
         case XDG_TOPLEVEL_STATE_MAXIMIZED:    is_maximized = true; break;
+        case XDG_TOPLEVEL_STATE_FULLSCREEN:   is_fullscreen = true; break;
+        case XDG_TOPLEVEL_STATE_RESIZING:     is_resizing = true; break;
+        case XDG_TOPLEVEL_STATE_ACTIVATED:    is_activated = true; break;
         case XDG_TOPLEVEL_STATE_TILED_LEFT:   is_tiled_left = true; break;
         case XDG_TOPLEVEL_STATE_TILED_RIGHT:  is_tiled_right = true; break;
         case XDG_TOPLEVEL_STATE_TILED_TOP:    is_tiled_top = true; break;
         case XDG_TOPLEVEL_STATE_TILED_BOTTOM: is_tiled_bottom = true; break;
-        case XDG_TOPLEVEL_STATE_RESIZING:     is_resizing = true; break;
-        }
+
+#if defined(XDG_TOPLEVEL_STATE_SUSPENDED_SINCE_VERSION)
+        case XDG_TOPLEVEL_STATE_SUSPENDED:    is_suspended = true; break;
+#endif
+    }
 
 #if defined(LOG_ENABLE_DBG) && LOG_ENABLE_DBG
-        if (*state >= XDG_TOPLEVEL_STATE_MAXIMIZED &&
-            *state <= XDG_TOPLEVEL_STATE_TILED_BOTTOM)
-        {
+        if (*state >= 0 && *state < ALEN(strings)) {
             state_chars += snprintf(
                 &state_str[state_chars], sizeof(state_str) - state_chars,
-                "%s, ", strings[*state]);
+                "%s, ",
+                strings[*state] != NULL ? strings[*state] : "<unknown>");
         }
 #endif
     }
@@ -848,7 +949,7 @@ xdg_surface_configure(void *data, struct xdg_surface *xdg_surface,
          * anytime soon. Some compositors require a commit in
          * combination with an ack - make them happy.
          */
-        wl_surface_commit(win->surface);
+        wl_surface_commit(win->surface.surf);
     }
 
     if (wasnt_configured)
@@ -1121,6 +1222,38 @@ handle_global(void *data, struct wl_registry *registry,
     }
 #endif
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+    else if (strcmp(interface, wp_viewporter_interface.name) == 0) {
+        const uint32_t required = 1;
+        if (!verify_iface_version(interface, version, required))
+            return;
+
+        wayl->viewporter = wl_registry_bind(
+            wayl->registry, name, &wp_viewporter_interface, required);
+    }
+
+    else if (strcmp(interface, wp_fractional_scale_manager_v1_interface.name) == 0) {
+        const uint32_t required = 1;
+        if (!verify_iface_version(interface, version, required))
+            return;
+
+        wayl->fractional_scale_manager = wl_registry_bind(
+            wayl->registry, name,
+            &wp_fractional_scale_manager_v1_interface, required);
+    }
+#endif
+
+#if defined(HAVE_CURSOR_SHAPE)
+    else if (strcmp(interface, wp_cursor_shape_manager_v1_interface.name) == 0) {
+        const uint32_t required = 1;
+        if (!verify_iface_version(interface, version, required))
+            return;
+
+        wayl->cursor_shape_manager = wl_registry_bind(
+            wayl->registry, name, &wp_cursor_shape_manager_v1_interface, required);
+    }
+#endif
+
 #if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
     else if (strcmp(interface, zwp_text_input_manager_v3_interface.name) == 0) {
         const uint32_t required = 1;
@@ -1204,7 +1337,7 @@ handle_global_remove(void *data, struct wl_registry *registry, uint32_t name)
 
             if (seat->wl_keyboard != NULL)
                 keyboard_listener.leave(
-                    seat, seat->wl_keyboard, -1, seat->kbd_focus->window->surface);
+                    seat, seat->wl_keyboard, -1, seat->kbd_focus->window->surface.surf);
         }
 
         if (seat->mouse_focus != NULL) {
@@ -1214,7 +1347,7 @@ handle_global_remove(void *data, struct wl_registry *registry, uint32_t name)
 
             if (seat->wl_pointer != NULL)
                 pointer_listener.leave(
-                    seat, seat->wl_pointer, -1, seat->mouse_focus->window->surface);
+                    seat, seat->wl_pointer, -1, seat->mouse_focus->window->surface.surf);
         }
 
         seat_destroy(seat);
@@ -1334,6 +1467,11 @@ wayl_init(struct fdm *fdm, struct key_binding_manager *key_binding_manager,
         LOG_ERR("no seats available (wl_seat interface too old?)");
         goto out;
     }
+    if (tll_length(wayl->monitors) == 0) {
+        LOG_ERR("no monitors available");
+        goto out;
+    }
+
     if (wayl->primary_selection_device_manager == NULL)
         LOG_WARN("no primary selection available");
 
@@ -1347,6 +1485,23 @@ wayl_init(struct fdm *fdm, struct key_binding_manager *key_binding_manager,
             "bell.urgent will fall back to coloring the window margins red");
     }
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (wayl->fractional_scale_manager == NULL || wayl->viewporter == NULL) {
+#else
+    if (true) {
+#endif
+        LOG_WARN("fractional scaling not available");
+    }
+
+#if defined(HAVE_CURSOR_SHAPE)
+    if (wayl->cursor_shape_manager == NULL) {
+#else
+    if (true) {
+#endif
+        LOG_WARN("no server-side cursors available, "
+                 "falling back to client-side cursors");
+    }
+
     if (presentation_timings && wayl->presentation == NULL) {
         LOG_ERR("presentation time interface not implemented by compositor");
         goto out;
@@ -1369,14 +1524,12 @@ wayl_init(struct fdm *fdm, struct key_binding_manager *key_binding_manager,
 
     tll_foreach(wayl->monitors, it) {
         LOG_INFO(
-            "%s: %dx%d+%dx%d@%dHz %s %.2f\" scale=%d PPI=%dx%d (physical) PPI=%dx%d (logical), DPI=%.2f",
+            "%s: %dx%d+%dx%d@%dHz %s %.2f\" scale=%d, DPI=%.2f/%.2f (physical/scaled)",
             it->item.name, it->item.dim.px_real.width, it->item.dim.px_real.height,
             it->item.x, it->item.y, (int)round(it->item.refresh),
             it->item.model != NULL ? it->item.model : it->item.description,
             it->item.inch, it->item.scale,
-            it->item.ppi.real.x, it->item.ppi.real.y,
-            it->item.ppi.scaled.x, it->item.ppi.scaled.y,
-            it->item.dpi);
+            it->item.dpi.physical, it->item.dpi.scaled);
     }
 
     wayl->fd = wl_display_get_fd(wayl->display);
@@ -1435,6 +1588,16 @@ wayl_destroy(struct wayland *wayl)
         zwp_text_input_manager_v3_destroy(wayl->text_input_manager);
 #endif
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (wayl->fractional_scale_manager != NULL)
+        wp_fractional_scale_manager_v1_destroy(wayl->fractional_scale_manager);
+    if (wayl->viewporter != NULL)
+        wp_viewporter_destroy(wayl->viewporter);
+#endif
+#if defined(HAVE_CURSOR_SHAPE)
+    if (wayl->cursor_shape_manager != NULL)
+        wp_cursor_shape_manager_v1_destroy(wayl->cursor_shape_manager);
+#endif
 #if defined(HAVE_XDG_ACTIVATION)
     if (wayl->xdg_activation != NULL)
         xdg_activation_v1_destroy(wayl->xdg_activation);
@@ -1469,6 +1632,29 @@ wayl_destroy(struct wayland *wayl)
     free(wayl);
 }
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+static void fractional_scale_preferred_scale(
+    void *data, struct wp_fractional_scale_v1 *wp_fractional_scale_v1,
+    uint32_t scale)
+{
+    struct wl_window *win = data;
+
+    const float new_scale = (float)scale / 120.;
+
+    if (win->scale == new_scale)
+        return;
+
+    LOG_DBG("fractional scale: %.2f -> %.2f", win->scale, new_scale);
+
+    win->scale = new_scale;
+    update_term_for_output_change(win->term);
+}
+
+static const struct wp_fractional_scale_v1_listener fractional_scale_listener = {
+    .preferred_scale = &fractional_scale_preferred_scale,
+};
+#endif
+
 struct wl_window *
 wayl_win_init(struct terminal *term, const char *token)
 {
@@ -1485,30 +1671,34 @@ wayl_win_init(struct terminal *term, const char *token)
     win->csd_mode = CSD_UNKNOWN;
     win->csd.move_timeout_fd = -1;
     win->resize_timeout_fd = -1;
+    win->scale = -1.;
 
     win->wm_capabilities.maximize = true;
     win->wm_capabilities.minimize = true;
 
-    win->surface = wl_compositor_create_surface(wayl->compositor);
-    if (win->surface == NULL) {
+    win->surface.surf = wl_compositor_create_surface(wayl->compositor);
+    if (win->surface.surf == NULL) {
         LOG_ERR("failed to create wayland surface");
         goto out;
     }
 
-    if (term->colors.alpha == 0xffff) {
-        struct wl_region *region = wl_compositor_create_region(
-            term->wl->compositor);
+    wayl_win_alpha_changed(win);
 
-        if (region != NULL) {
-            wl_region_add(region, 0, 0, INT32_MAX, INT32_MAX);
-            wl_surface_set_opaque_region(win->surface, region);
-            wl_region_destroy(region);
-        }
+    wl_surface_add_listener(win->surface.surf, &surface_listener, win);
+
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (wayl->fractional_scale_manager != NULL && wayl->viewporter != NULL) {
+        win->surface.viewport = wp_viewporter_get_viewport(wayl->viewporter, win->surface.surf);
+
+        win->fractional_scale =
+            wp_fractional_scale_manager_v1_get_fractional_scale(
+                wayl->fractional_scale_manager, win->surface.surf);
+        wp_fractional_scale_v1_add_listener(
+            win->fractional_scale, &fractional_scale_listener, win);
     }
+#endif
 
-    wl_surface_add_listener(win->surface, &surface_listener, win);
-
-    win->xdg_surface = xdg_wm_base_get_xdg_surface(wayl->shell, win->surface);
+    win->xdg_surface = xdg_wm_base_get_xdg_surface(wayl->shell, win->surface.surf);
     xdg_surface_add_listener(win->xdg_surface, &xdg_surface_listener, win);
 
     win->xdg_toplevel = xdg_surface_get_toplevel(win->xdg_surface);
@@ -1541,12 +1731,12 @@ wayl_win_init(struct terminal *term, const char *token)
         LOG_WARN("no decoration manager available - using CSDs unconditionally");
     }
 
-    wl_surface_commit(win->surface);
+    wl_surface_commit(win->surface.surf);
 
 #if defined(HAVE_XDG_ACTIVATION)
     /* Complete XDG startup notification */
     if (token)
-        xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
+        xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface.surf);
 #endif
 
     if (!wayl_win_subsurface_new(win, &win->overlay, false)) {
@@ -1596,33 +1786,33 @@ wayl_win_destroy(struct wl_window *win)
      * nor mouse focus).
      */
 
-    if (win->render_timer.surf != NULL) {
-        wl_surface_attach(win->render_timer.surf, NULL, 0, 0);
-        wl_surface_commit(win->render_timer.surf);
+    if (win->render_timer.surface.surf != NULL) {
+        wl_surface_attach(win->render_timer.surface.surf, NULL, 0, 0);
+        wl_surface_commit(win->render_timer.surface.surf);
     }
 
-    if (win->scrollback_indicator.surf != NULL) {
-        wl_surface_attach(win->scrollback_indicator.surf, NULL, 0, 0);
-        wl_surface_commit(win->scrollback_indicator.surf);
+    if (win->scrollback_indicator.surface.surf != NULL) {
+        wl_surface_attach(win->scrollback_indicator.surface.surf, NULL, 0, 0);
+        wl_surface_commit(win->scrollback_indicator.surface.surf);
     }
 
     /* Scrollback search */
-    if (win->search.surf != NULL) {
-        wl_surface_attach(win->search.surf, NULL, 0, 0);
-        wl_surface_commit(win->search.surf);
+    if (win->search.surface.surf != NULL) {
+        wl_surface_attach(win->search.surface.surf, NULL, 0, 0);
+        wl_surface_commit(win->search.surface.surf);
     }
 
     /* URLs */
     tll_foreach(win->urls, it) {
-        wl_surface_attach(it->item.surf.surf, NULL, 0, 0);
-        wl_surface_commit(it->item.surf.surf);
+        wl_surface_attach(it->item.surf.surface.surf, NULL, 0, 0);
+        wl_surface_commit(it->item.surf.surface.surf);
     }
 
     /* CSD */
     for (size_t i = 0; i < ALEN(win->csd.surface); i++) {
-        if (win->csd.surface[i].surf != NULL) {
-            wl_surface_attach(win->csd.surface[i].surf, NULL, 0, 0);
-            wl_surface_commit(win->csd.surface[i].surf);
+        if (win->csd.surface[i].surface.surf != NULL) {
+            wl_surface_attach(win->csd.surface[i].surface.surf, NULL, 0, 0);
+            wl_surface_commit(win->csd.surface[i].surface.surf);
         }
     }
 
@@ -1630,8 +1820,8 @@ wayl_win_destroy(struct wl_window *win)
 
         /* Main window */
     win->unmapped = true;
-    wl_surface_attach(win->surface, NULL, 0, 0);
-    wl_surface_commit(win->surface);
+    wl_surface_attach(win->surface.surf, NULL, 0, 0);
+    wl_surface_commit(win->surface.surf);
     wayl_roundtrip(win->term->wl);
 
     tll_free(win->on_outputs);
@@ -1661,6 +1851,12 @@ wayl_win_destroy(struct wl_window *win)
 
         tll_remove(win->xdg_tokens, it);
     }
+#endif
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (win->fractional_scale != NULL)
+        wp_fractional_scale_v1_destroy(win->fractional_scale);
+    if (win->surface.viewport != NULL)
+        wp_viewport_destroy(win->surface.viewport);
 #endif
     if (win->frame_callback != NULL)
         wl_callback_destroy(win->frame_callback);
@@ -1670,8 +1866,8 @@ wayl_win_destroy(struct wl_window *win)
         xdg_toplevel_destroy(win->xdg_toplevel);
     if (win->xdg_surface != NULL)
         xdg_surface_destroy(win->xdg_surface);
-    if (win->surface != NULL)
-        wl_surface_destroy(win->surface);
+    if (win->surface.surf != NULL)
+        wl_surface_destroy(win->surface.surf);
 
     wayl_roundtrip(win->term->wl);
 
@@ -1681,7 +1877,7 @@ wayl_win_destroy(struct wl_window *win)
 }
 
 bool
-wayl_reload_xcursor_theme(struct seat *seat, int new_scale)
+wayl_reload_xcursor_theme(struct seat *seat, float new_scale)
 {
     if (seat->pointer.theme != NULL && seat->pointer.scale == new_scale) {
         /* We already have a theme loaded, and the scale hasn't changed */
@@ -1714,7 +1910,7 @@ wayl_reload_xcursor_theme(struct seat *seat, int new_scale)
 
     const char *xcursor_theme = getenv("XCURSOR_THEME");
 
-    LOG_INFO("cursor theme: %s, size: %d, scale: %d",
+    LOG_INFO("cursor theme: %s, size: %d, scale: %.2f",
              xcursor_theme ? xcursor_theme : "(null)",
              xcursor_size, new_scale);
 
@@ -1798,6 +1994,86 @@ wayl_roundtrip(struct wayland *wayl)
     wayl_flush(wayl);
 }
 
+
+bool
+wayl_fractional_scaling(const struct wayland *wayl)
+{
+#if defined(HAVE_FRACTIONAL_SCALE)
+    return wayl->fractional_scale_manager != NULL;
+#else
+    return false;
+#endif
+}
+
+void
+wayl_surface_scale_explicit_width_height(
+    const struct wl_window *win, const struct wayl_surface *surf,
+    int width, int height, float scale)
+{
+
+    if (wayl_fractional_scaling(win->term->wl) && win->scale > 0.) {
+#if defined(HAVE_FRACTIONAL_SCALE)
+        LOG_DBG("scaling by a factor of %.2f using fractional scaling "
+                "(width=%d, height=%d) ", scale, width, height);
+
+        wl_surface_set_buffer_scale(surf->surf, 1);
+        wp_viewport_set_destination(
+            surf->viewport,
+            round((float)width / scale),
+            round((float)height / scale));
+#else
+        BUG("wayl_fraction_scaling() returned true, "
+            "but fractional scaling was not available at compile time");
+#endif
+    } else {
+        LOG_DBG("scaling by a factor of %.2f using legacy mode "
+                "(width=%d, height=%d)", scale, width, height);
+
+        xassert(scale == floor(scale));
+
+        const int iscale = (int)scale;
+        xassert(width % iscale == 0);
+        xassert(height % iscale == 0);
+
+        wl_surface_set_buffer_scale(surf->surf, iscale);
+    }
+}
+
+void
+wayl_surface_scale(const struct wl_window *win, const struct wayl_surface *surf,
+                   const struct buffer *buf, float scale)
+{
+    wayl_surface_scale_explicit_width_height(
+        win, surf, buf->width, buf->height, scale);
+}
+
+void
+wayl_win_scale(struct wl_window *win, const struct buffer *buf)
+{
+    const struct terminal *term = win->term;
+    const float scale = term->scale;
+
+    wayl_surface_scale(win, &win->surface, buf, scale);
+}
+
+void
+wayl_win_alpha_changed(struct wl_window *win)
+{
+    struct terminal *term = win->term;
+
+    if (term->colors.alpha == 0xffff) {
+        struct wl_region *region = wl_compositor_create_region(
+            term->wl->compositor);
+
+        if (region != NULL) {
+            wl_region_add(region, 0, 0, INT32_MAX, INT32_MAX);
+            wl_surface_set_opaque_region(win->surface.surf, region);
+            wl_region_destroy(region);
+        }
+    } else
+        wl_surface_set_opaque_region(win->surface.surf, NULL);
+}
+
 #if defined(HAVE_XDG_ACTIVATION)
 static void
 activation_token_for_urgency_done(const char *token, void *data)
@@ -1806,7 +2082,7 @@ activation_token_for_urgency_done(const char *token, void *data)
     struct wayland *wayl = win->term->wl;
 
     win->urgency_token_is_pending = false;
-    xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
+    xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface.surf);
 }
 #endif /* HAVE_XDG_ACTIVATION */
 
@@ -1851,27 +2127,43 @@ wayl_win_csd_borders_visible(const struct wl_window *win)
 bool
 wayl_win_subsurface_new_with_custom_parent(
     struct wl_window *win, struct wl_surface *parent,
-    struct wl_surf_subsurf *surf, bool allow_pointer_input)
+    struct wayl_sub_surface *surf, bool allow_pointer_input)
 {
     struct wayland *wayl = win->term->wl;
 
-    surf->surf = NULL;
+    surf->surface.surf = NULL;
     surf->sub = NULL;
 
     struct wl_surface *main_surface
         = wl_compositor_create_surface(wayl->compositor);
 
-    if (main_surface == NULL)
+    if (main_surface == NULL) {
+        LOG_ERR("failed to instantiate surface for sub-surface");
         return false;
+    }
 
     struct wl_subsurface *sub = wl_subcompositor_get_subsurface(
         wayl->sub_compositor, main_surface, parent);
 
     if (sub == NULL) {
+        LOG_ERR("failed to instantiate sub-surface");
         wl_surface_destroy(main_surface);
         return false;
     }
 
+#if defined(HAVE_FRACTIONAL_SCALE)
+    struct wp_viewport *viewport = NULL;
+    if (wayl->fractional_scale_manager != NULL &&  wayl->viewporter != NULL) {
+        viewport = wp_viewporter_get_viewport(wayl->viewporter, main_surface);
+        if (viewport == NULL) {
+            LOG_ERR("failed to instantiate viewport for sub-surface");
+            wl_subsurface_destroy(sub);
+            wl_surface_destroy(main_surface);
+            return false;
+        }
+    }
+#endif
+
     wl_surface_set_user_data(main_surface, win);
     wl_subsurface_set_sync(sub);
 
@@ -1883,31 +2175,42 @@ wayl_win_subsurface_new_with_custom_parent(
         wl_region_destroy(empty);
     }
 
-    surf->surf = main_surface;
+    surf->surface.surf = main_surface;
     surf->sub = sub;
+#if defined(HAVE_FRACTIONAL_SCALE)
+    surf->surface.viewport = viewport;
+#endif
     return true;
 }
 
 bool
-wayl_win_subsurface_new(struct wl_window *win, struct wl_surf_subsurf *surf,
+wayl_win_subsurface_new(struct wl_window *win, struct wayl_sub_surface *surf,
                         bool allow_pointer_input)
 {
     return wayl_win_subsurface_new_with_custom_parent(
-        win, win->surface, surf, allow_pointer_input);
+        win, win->surface.surf, surf, allow_pointer_input);
 }
 
 void
-wayl_win_subsurface_destroy(struct wl_surf_subsurf *surf)
+wayl_win_subsurface_destroy(struct wayl_sub_surface *surf)
 {
     if (surf == NULL)
         return;
-    if (surf->sub != NULL)
-        wl_subsurface_destroy(surf->sub);
-    if (surf->surf != NULL)
-        wl_surface_destroy(surf->surf);
 
-    surf->surf = NULL;
-    surf->sub = NULL;
+#if defined(HAVE_FRACTIONAL_SCALE)
+    if (surf->surface.viewport != NULL) {
+        wp_viewport_destroy(surf->surface.viewport);
+        surf->surface.viewport = NULL;
+    }
+#endif
+    if (surf->sub != NULL) {
+        wl_subsurface_destroy(surf->sub);
+        surf->sub = NULL;
+    }
+    if (surf->surface.surf != NULL) {
+        wl_surface_destroy(surf->surface.surf);
+        surf->surface.surf = NULL;
+    }
 }
 
 #if defined(HAVE_XDG_ACTIVATION)
@@ -1972,7 +2275,7 @@ wayl_get_activation_token(
     if (seat != NULL && serial != 0)
         xdg_activation_token_v1_set_serial(token, serial, seat->wl_seat);
 
-    xdg_activation_token_v1_set_surface(token, win->surface);
+    xdg_activation_token_v1_set_surface(token, win->surface.surf);
     xdg_activation_token_v1_add_listener(token, &activation_token_listener, ctx);
     xdg_activation_token_v1_commit(token);
     return true;
